from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import pickle
import xml.etree.ElementTree

import networkx as nx
import nltk
import numpy as np


class DBLP(object):
    @staticmethod
    def get_articles(dir_path='./'):
        pkl_path = os.path.join(dir_path, 'dblp_cs.pkl')
        if not os.path.exists(pkl_path):
            articles = DBLP._extract_cs(dir_path)
            with open(pkl_path, 'wb') as fout:
                pickle.dump(articles, fout)
        else:
            with open(pkl_path, 'rb') as fin:
                articles = pickle.load(fin)
        return articles

    @staticmethod
    def _extract_cs(dir_path):
        xml_path = os.path.join(dir_path, 'dblp.xml')
        root = xml.etree.ElementTree.parse(xml_path).getroot()
        articles = []
        for venue in ['BIBM', 'RECOMB',
                      'INFOCOM', 'SIGCOMM',
                      'SC', 'ISCA',
                      'EUROCRYPT', 'CRYPTO',
                      'DCC',
                      'CVPR', 'ICCV',
                      'ACL', 'COLING',
                      'SIGMOD Conference', 'VLDB']:
            cnt = 0
            for paper in root.iter('inproceedings'):
                book = next(paper.iter('booktitle')).text
                year = int(next(paper.iter('year')).text)
                if book == venue and year <= 2016:
                    articles.append(paper)
                    cnt += 1
            print('Added %d papers from %s.' % (cnt, venue))
        print('Extracted %d papers in total.' % len(articles))
        return articles


def main_title(dir_path='./'):
    # Read and tokenize titles.
    wrd_freq = {}
    venues = {}
    for paper in DBLP.get_articles(dir_path):
        tl = next(paper.iter('title')).text
        if tl is None:
            continue
        tl = nltk.word_tokenize(tl)
        tl = [w.lower() for w in tl]
        for w in tl:
            if w in wrd_freq:
                wrd_freq[w] += 1
            else:
                wrd_freq[w] = 0
        vn = next(paper.iter('booktitle')).text
        if vn not in venues:
            venues[vn] = []
        venues[vn].append(tl)

    # Remove rare words.
    vocab = set([w for w in wrd_freq if wrd_freq[w] > 1])
    print(len(vocab))
    for v in venues:
        tmp = []
        for tl in venues[v]:
            tl = [w for w in tl if w in vocab]
            if len(tl) > 0:
                tmp.append(tl)
        venues[v] = tmp
    for v in venues:
        assert len(venues[v]) > 0

    # Compute tf-idf.
    tf = dict((v, {}) for v in venues)
    for v in venues:
        for tl in venues[v]:
            for w in tl:
                if w in tf[v]:
                    tf[v][w] += 1
                else:
                    tf[v][w] = 1
    idf = dict((w, 0) for w in vocab)
    for w in vocab:
        cnt = 0
        for v in venues:
            if w in tf[v]:
                assert tf[v][w] > 0
                cnt += 1
        assert cnt > 0
        idf[w] = np.log(len(venues) / cnt)
    tf_idf = dict((v, {}) for v in venues)
    for v in venues:
        for tl in venues[v]:
            for w in tl:
                tf_idf[v][w] = tf[v][w] * idf[w]

    # Construct the graph.
    g = nx.Graph()
    for v in venues:
        scores = sorted(tf_idf[v].values())
        assert len(scores) > 0
        thresh = scores[-int(0.05 * len(scores))]
        for tl in venues[v]:
            for x in tl:
                if tf_idf[v][x] > thresh:
                    for y in tl:
                        if x != y and tf_idf[v][y] > thresh:
                            g.add_edge(x, y)

    # Refine and save the graph.
    print(g.number_of_nodes(), g.number_of_edges())
    g = next(nx.connected_component_subgraphs(g))
    print(g.number_of_nodes(), g.number_of_edges())
    nx.write_gml(g, 'db.gml')
    mapping = dict(zip(g, range(g.number_of_nodes())))
    with open('db.voc', 'w') as fout:
        for w, i in mapping.items():
            fout.write('%d %s\n' % (i, w))
    g = nx.relabel_nodes(g, mapping)
    nx.write_edgelist(g, 'db.elist', data=False)


if __name__ == '__main__':
    np.random.seed(0)
    main_title()
